/*
 * $Id: file.c,v 1.14 2000/02/06 10:37:46 tatuo-y Exp $
 *
 * open/close files
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>

#include "sufary.h"
#include "util.h"
#include "config.h"


static SA_STAT sa_opentextfile(SUFARY *ary, const char *s);
static SA_STAT sa_openarrayfile(SUFARY *ary, const char *s);
static void sa_closetextfile(SUFARY *ary);
static void sa_closearrayfile(SUFARY *ary);


/*
 * access functions
 */
char *
sa_get_text_ptr(const SUFARY *ary)
{
    return ary->text;
}

SA_INDEX
sa_get_text_size(const SUFARY *ary)
{
    return ary->textsize;
}

SA_INDEX *
sa_get_array_ptr(const SUFARY *ary)
{
    return ary->array;
}

SA_INDEX
sa_get_array_size(const SUFARY *ary)
{
    return ary->arraysize;
}

SA_INDEX *
sa_get_did_ptr(const DID *did)
{
    return did->docid;
}

SA_INDEX
sa_get_did_size(const DID *did)
{
    /* number of documents, not number of indexes */
    return did->didsize / 2;
}


/*
 *   ꤵ줿ƥȥեarrayե򳫤
 *   arrayե̾NULL˻ꤹСƥȥե̾
 *   '.ary' ղäΤarrayե̾ˤʤ롣
 */
SUFARY *
sa_open(const char *txt_fname, char *ary_fname)
{
    SUFARY *newary;
    char fname_tmp[SA_FILE_NAME_MAX];

    if (ary_fname == NULL || ary_fname[0] == '\0')
	ary_fname = sa_add_suffix_to_file_name(fname_tmp, txt_fname, "ary");

    newary = (SUFARY *)calloc(sizeof(SUFARY), 1);
    if (!newary) {
	sa_error_no = MEMORY_ALLOCATE_ERROR;
	return NULL;
    }

    if (sa_opentextfile(newary, txt_fname) == FAIL ||
	sa_openarrayfile(newary, ary_fname) == FAIL) {
	sa_close(newary);
	free(newary);
	sa_error_no = FILE_OPEN_ERROR;
	return NULL;
    }

    return newary;
}


/*
 * open only a text file
 */
static SA_STAT
sa_opentextfile(SUFARY *ary, const char *filename)
{
    SA_MMAP *mms;

    mms = sa_open_mmap(filename, SA_MMAP_RO);
    if (mms == NULL) {
	sa_error_no = MMAP_ERROR;
	return FAIL;
    }

    ary->mmtxt = mms;

    ary->textsize  = sa_get_mmap_size(mms);
    ary->text = sa_get_mmap_ptr(mms);

    return SUCCESS;
}


/*
 * open only an array file
 */
static SA_STAT
sa_openarrayfile(SUFARY *ary, const char *filename)
{
    SA_MMAP *mms;

    mms = sa_open_mmap(filename, SA_MMAP_RO);
    if (mms == NULL) {
	sa_error_no = MMAP_ERROR;
	return FAIL;
    }

    ary->mmary = mms;

    ary->arraysize = sa_get_mmap_size(mms) / sizeof(SA_INDEX);
    ary->array = (SA_INDEX*)sa_get_mmap_ptr(mms);

    return SUCCESS;
}


/*
 * close a text file and an array file
 */
void
sa_close(SUFARY *ary)
{
    sa_closetextfile(ary);
    sa_closearrayfile(ary);
    free(ary);
    ary = NULL;
}


/*
 * close only a text file
 */
void
sa_closetextfile(SUFARY *ary)
{
    sa_close_mmap(ary->mmtxt);
}


/*
 * close only an array file
 */
void
sa_closearrayfile(SUFARY *ary)
{
    sa_close_mmap(ary->mmary);

    /* free(ary) ɬפ, Perl Ȥη͹礤ΤǤޤ٥å */
}


/*
 * open DocID file
 */
DID *
sa_open_did(const char *filename)
{
    DID *newdid;
    SA_MMAP *mms;

    newdid = (DID *)calloc(sizeof(DID), 1);
    if (!newdid){
	sa_error_no = MEMORY_ALLOCATE_ERROR;
	return NULL;
    }

    mms = sa_open_mmap(filename, SA_MMAP_RO);
    if (mms == NULL) {
	sa_error_no = MMAP_ERROR;
	return NULL;
    }

    newdid->mmdid = mms;

    newdid->docid = (SA_INDEX*)sa_get_mmap_ptr(mms);
    newdid->didsize = sa_get_mmap_size(mms) / sizeof(SA_INDEX);
    return newdid;
}


/*
 * close DocID file
 */
void
sa_close_did(DID *d)
{
    sa_close_mmap(d->mmdid);

/*     munmap(d->didmap, d->didsz); */
/*     close(d->didfd); */
/*     d->didmap = NULL; */
/*     d->didfd = -1; */
/*     d->didsz = 0; */

    /* free(did) ɬפ, Perl Ȥη͹礤ΤǤޤ٥å */
}


/*
 * write a big endian ordering long int
 */
size_t
sa_fwrite(SA_INDEX idx, FILE *fd)
{
#ifndef WORDS_BIGENDIAN
    sa_reverse_byte_order(&idx, sizeof(SA_INDEX));
#endif
    return fwrite(&idx, 1, sizeof(SA_INDEX), fd);
}



/* 
 * access functions
 */
off_t
sa_get_mmap_size(const SA_MMAP *mmr)
{
    return mmr->sz;
}

caddr_t
sa_get_mmap_ptr(const SA_MMAP *mmr)
{
    return mmr->map;
}


/*
 * open and mmap a file
 */
SA_MMAP *
sa_open_mmap(const char *fname, const SA_MMAP_MODE mode)
{
    SA_MMAP* mm;
    int fd;
    struct stat stat_buf;
    int oflag;
    int prot;

    mm = malloc(sizeof(SA_MMAP));
    if (mm == NULL)
	return NULL;

    /* set parameters */
    if (mode == SA_MMAP_RO) {
	oflag = O_RDONLY;
	prot = PROT_READ;
    } else if (mode == SA_MMAP_RW) {
	oflag = O_RDWR;
	prot = PROT_READ | PROT_WRITE;
    } else {
	return NULL;
    }
    
    /* open the addressed file */
    fd = open(fname, oflag);
    if (fd == -1) {
	return NULL;
    }

    /* size of the file */
    if (fstat(fd, &stat_buf) != 0) {
      close(fd);
      return NULL;
    }
    mm->sz = stat_buf.st_size;

    /* ready to load the file */
    mm->map = mmap((caddr_t)0, mm->sz, prot, MAP_SHARED, fd, 0);
    close(fd);
    if (mm->map == (caddr_t) -1) {
	return NULL;
    }

    return mm;
}


/*
 * un-mmap and close a file
 */
void
sa_close_mmap(SA_MMAP *mms)
{
    if (mms != NULL) {
	munmap(mms->map, mms->sz);
	free(mms);
    }
}

