/*
 * $Id: make-index.c,v 1.19 2000/02/06 10:37:46 tatuo-y Exp $
 *
 * make indexes
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <assert.h>

#include "sufary.h"
#include "util.h"
#include "config.h"

#define STR_CACHE_SIZE 20
/* for heap */
typedef struct {
    SA_INDEX aryidx;
    SA_INDEX max;
    char strbuf[STR_CACHE_SIZE + 1];
} BLOCK;

static void insertheap(BLOCK** a, BLOCK* v);
static void upheap(BLOCK** a, int k);
static void downheap(BLOCK** a, int k);
static BLOCK* hremove(BLOCK** a);

static int m_size_of_heap = 0;
static char* m_text_ptr;
static SA_INDEX* m_suffix;	/* used in heap */

static SA_STAT sa_sort_each_block(SA_INDEX, SA_INDEX, BLOCK**);
static SA_STAT sa_merge_blocks(char*, SA_INDEX, BLOCK**);

static char *ignored_chars = NULL;
static char *my_chars = NULL;
static char *my_delimiters = NULL;


/*
 * access functions
 */
void
sa_set_ignored_chars(char *p)
{
    ignored_chars = p;
}

char *
sa_get_ignored_chars(void)
{
    return ignored_chars;
}

void
sa_set_my_chars(char *p)
{
    my_chars = p;
}

char *
sa_get_my_chars(void)
{
    return my_chars;
}

void
sa_set_my_delimiters(char *p)
{
    my_delimiters = p;
}

char *
sa_get_my_delimiters(void)
{
    return my_delimiters;
}


/*
 * Compare suffixes for sort.
 */
static int 
suffix_compare(const SA_INDEX *a, const SA_INDEX *b)
{
    
#ifdef WORDS_BIGENDIAN
    return strcmp(m_text_ptr + *a, m_text_ptr + *b);
#else
    SA_INDEX aa = *a;
    SA_INDEX bb = *b;
    sa_reverse_byte_order(&aa, sizeof(SA_INDEX));
    sa_reverse_byte_order(&bb, sizeof(SA_INDEX));
    return strcmp(m_text_ptr + aa, m_text_ptr + bb);
#endif
}


/*
 * Sort index-pointers.
 */
SA_STAT
sa_sort_index(const char *txt_fname, char *ary_fname)
{
    SA_MMAP *mmtxt, *mmary;
    char fname_tmp[SA_FILE_NAME_MAX];

    if (ary_fname == NULL || ary_fname[0] == '\0')
	ary_fname = sa_add_suffix_to_file_name(fname_tmp, txt_fname, "ary");

    mmtxt = sa_open_mmap(txt_fname, SA_MMAP_RO);
    mmary = sa_open_mmap(ary_fname, SA_MMAP_RW);

    if (mmtxt == NULL || mmary == NULL) {
	sa_error_no = MMAP_ERROR;
	return FAIL;
    }

    m_text_ptr = sa_get_mmap_ptr(mmtxt);

    {
	SA_INDEX *suf = (SA_INDEX*)sa_get_mmap_ptr(mmary);
	SA_INDEX num_of_indexes = sa_get_mmap_size(mmary) / sizeof(SA_INDEX);

	qsort(suf, (size_t)num_of_indexes, sizeof(SA_INDEX),
	      (int (*)(const void *, const void *))suffix_compare);
    }
	
    sa_close_mmap(mmary);
    sa_close_mmap(mmtxt);

    return SUCCESS;
}


/*
 * Sort index-pointers --- divide and merge
 */
SA_STAT
sa_sort_index_block(const char *txt_fname, char *ary_fname, int bl_size_mega)
{
    SA_MMAP *mmtxt, *mmary;
    char fname_tmp[SA_FILE_NAME_MAX];

    if (ary_fname == NULL || ary_fname[0] == '\0')
	ary_fname = sa_add_suffix_to_file_name(fname_tmp, txt_fname, "ary");

    mmtxt = sa_open_mmap(txt_fname, SA_MMAP_RO);
    mmary = sa_open_mmap(ary_fname, SA_MMAP_RW);

    if (mmtxt == NULL || mmary == NULL) {
	sa_error_no = MMAP_ERROR;
	return FAIL;
    }

    m_text_ptr = sa_get_mmap_ptr(mmtxt);

    {
	SA_INDEX num_of_indexes = sa_get_mmap_size(mmary) / sizeof(SA_INDEX);
	SA_INDEX bl_size = 1000000 / sizeof(SA_INDEX) * bl_size_mega;
	BLOCK **bl;
	char merged_ary_fname[SA_FILE_NAME_MAX];
	int num_of_blocks;

	num_of_blocks = num_of_indexes / bl_size;
	if (num_of_indexes % bl_size)
	    num_of_blocks++;

	bl = malloc((num_of_blocks + 1) * sizeof(BLOCK*)); 

	m_suffix = (SA_INDEX*)sa_get_mmap_ptr(mmary);

	sa_dprintf("sort blocks: num index = %ld, num block = %ld\n",
		   num_of_indexes, num_of_blocks);

	if (sa_sort_each_block(bl_size, num_of_indexes, bl) == FAIL)
	    return FAIL;
	
	sprintf(merged_ary_fname, "%s-", ary_fname);

	if (sa_merge_blocks(merged_ary_fname,
			    num_of_indexes, bl) == FAIL)
	    return FAIL;

	sa_dprintf("sort blocks: rename \"%s\" with \"%s\"\n",
		   merged_ary_fname, ary_fname);

	if (rename(merged_ary_fname, ary_fname) != 0) {
	    sa_error_no = FILE_OPEN_ERROR;
	    return FAIL;
	}

	free(*bl);
    }

    sa_close_mmap(mmary);
    sa_close_mmap(mmtxt);

    return SUCCESS;
}


/*
 * Sort each block
 */
static SA_STAT
sa_sort_each_block(SA_INDEX bl_size, SA_INDEX num_of_indexes, BLOCK **bl)
{
    SA_INDEX j;

    sa_dprintf("sort each block: block size = %ld\n", bl_size);

    for (j = 0; j < num_of_indexes; j += bl_size) {
	BLOCK* tmp = malloc(sizeof(BLOCK));
	SA_INDEX size = bl_size;

	if (j + bl_size > num_of_indexes)
	    size = num_of_indexes - j;

	tmp->aryidx = j;
	tmp->max = tmp->aryidx + size;

	sa_dprintf("sort each block: from=%ld to=%ld size=%ld\n",
		   tmp->aryidx, tmp->max, size);

	qsort(m_suffix + tmp->aryidx, (size_t)size, sizeof(SA_INDEX),
	      (int (*)(const void *,const void *))suffix_compare);

	{
	    SA_INDEX aa = m_suffix[tmp->aryidx];
#ifndef WORDS_BIGENDIAN
	    sa_reverse_byte_order(&aa, sizeof(SA_INDEX));
#endif
	    strncpy(tmp->strbuf, m_text_ptr + aa, STR_CACHE_SIZE);
	}

	tmp->strbuf[STR_CACHE_SIZE] = '\0';
	insertheap(bl, tmp);

    }

    return SUCCESS;
}    


/*
 * Merge all blocks into an index file.
 */
static SA_STAT
sa_merge_blocks(char *merged_ary_fname, SA_INDEX num_of_indexes, BLOCK **bl)
{
    int i;
    FILE* mfd;

    sa_dprintf("merge index: start!\n");

    if ((mfd = fopen(merged_ary_fname, "wb")) == NULL) {
	sa_error_no = FILE_OPEN_ERROR;
	return FAIL;
    }

    for (i = 0; i < num_of_indexes; i++) {
	BLOCK* tmp = bl[1];

	fwrite((m_suffix + tmp->aryidx), 1, sizeof(SA_INDEX), mfd);

	(tmp->aryidx)++;
	if (tmp->aryidx >= tmp->max) {
	    hremove(bl);
	    continue;
	}

	{
	    SA_INDEX aa = m_suffix[tmp->aryidx];
#ifndef WORDS_BIGENDIAN
	    sa_reverse_byte_order(&aa, sizeof(SA_INDEX));
#endif
	    strncpy(tmp->strbuf, m_text_ptr + aa, STR_CACHE_SIZE);
	}

	tmp->strbuf[STR_CACHE_SIZE] = '\0';
	downheap(bl, 1);
    }
    
    fclose(mfd);

    return SUCCESS;
}


/*
 * Heap
 *   Reference: R. Sedgewick  ʿ顢ƣϡĸ , 
 *     "르ꥺ (Algorithms) 2 1 á", ʳؼ
 */
static int
two_level_compare(const BLOCK *a, const BLOCK *b)
{
    int r = strcmp(a->strbuf, b->strbuf);
    if (r == 0) {
	SA_INDEX aa = m_suffix[a->aryidx];
	SA_INDEX bb = m_suffix[b->aryidx];
#ifndef WORDS_BIGENDIAN
	sa_reverse_byte_order(&aa, sizeof(SA_INDEX));
	sa_reverse_byte_order(&bb, sizeof(SA_INDEX));
#endif
	return strcmp(m_text_ptr + aa + STR_CACHE_SIZE,
		      m_text_ptr + bb + STR_CACHE_SIZE);
    }
    return r;
}

static void
upheap(BLOCK **a, int k)
{
    BLOCK *v;
    v = a[k];
    while (k > 1 &&
	   suffix_compare(m_suffix + a[k / 2]->aryidx,
			  m_suffix + v->aryidx) > 0) {
	a[k] = a[k / 2];
	k = k / 2;
    }
    a[k] = v;
}

static void
insertheap(BLOCK **a, BLOCK *v)
{
    m_size_of_heap++;
    a[m_size_of_heap] = v;
    upheap(a, m_size_of_heap);
}

static void
downheap(BLOCK **a, int k)
{
    int j;
    BLOCK *v;
    v = a[k];
    while (k <= m_size_of_heap / 2) {
	j = k * 2;
	if (j < m_size_of_heap && two_level_compare(a[j], a[j + 1]) >= 0)
 	    j++;
	if (two_level_compare(v, a[j]) <= 0)
	    break;
	a[k] = a[j];
	k = j;
    }
    a[k] = v;
}

static BLOCK *
hremove(BLOCK **a)
{
    BLOCK *r = a[1];
    a[1] = a[m_size_of_heap];
    m_size_of_heap--;
    downheap(a, 1);
    return r;
}


/*
 * Scan a text file and write indexes to a array file
 */
SA_STAT
sa_write_index(const char *txt_fname, char *ary_fname, int mki_mode)
{
    SA_INDEX i;
    int pre_char_is_delimiter = 1;
    int pre_char_is_8th_bit_on = 0;
    FILE *ary_file_fd = NULL;
    int size_of_text_file;
    char* text;
    SA_MMAP *mmr;
    char fname_tmp[SA_FILE_NAME_MAX];

    if (ary_fname == NULL || ary_fname[0] == '\0')
	ary_fname = sa_add_suffix_to_file_name(fname_tmp, txt_fname, "ary");

    if ((ary_file_fd = fopen(ary_fname, "wb")) == NULL) {
	sa_error_no = FILE_OPEN_ERROR;
	return FAIL;
    }

    mmr = sa_open_mmap(txt_fname, SA_MMAP_RO);
    if (mmr == NULL) {
	sa_error_no = MMAP_ERROR;
	return FAIL;
    }

    size_of_text_file = sa_get_mmap_size(mmr);
    text = sa_get_mmap_ptr(mmr);


    if (mki_mode & SA_CUT_TOP) {
	pre_char_is_delimiter = 0;
	i = 1;
    } else
	i = 0;

    for (; i < size_of_text_file; i++) {

	if ((mki_mode & SA_VERBOSE) && i != 0) {
	    if (!(i % 50000))
		printf("+");
	    if (!(i % 1000000))
		printf(" %ldM\n", (long)i/1000000);
	}
	
	if (mki_mode & SA_AT) {
	    /* by char */

	    if (sa_get_my_chars() != NULL &&
		strchr(sa_get_my_chars(), text[i]) != NULL) {
		sa_fwrite(i, ary_file_fd);
	    } else if (sa_get_ignored_chars() != NULL &&
		       strchr(sa_get_ignored_chars(), text[i]) != NULL) {
		/* do nothing */
	    } else if ((mki_mode & SA_ONLY_00_7F) &&
		       (unsigned char)text[i] >= 0x80) {
		/* do nothing */
	    } else if ((mki_mode & SA_ONLY_80_FF) &&
		       (unsigned char)text[i] < 0x80) {
		/* do nothing */
	    } else if (mki_mode & SA_EUC_2) {

		if (MSB_ON(text[i]) && pre_char_is_8th_bit_on == 1) {
		    /* ignore a 2nd byte of 2 bytes char */
		    pre_char_is_8th_bit_on = 0;
		} else {
		    sa_fwrite(i, ary_file_fd);
		    if (MSB_ON(text[i]))
			pre_char_is_8th_bit_on = 1;
		}

	    } else
		sa_fwrite(i, ary_file_fd);

	} else if (mki_mode & SA_AFTER) {
	    /* after delimiter */

	    if (strchr(sa_get_my_delimiters(), text[i]) != NULL &&
		text[i] != '\0') {

		pre_char_is_delimiter = 1;

	    } else if (pre_char_is_delimiter == 1) {

		if (sa_get_my_chars() != NULL &&
		    strchr(sa_get_my_chars(), text[i]) != NULL) {
		    sa_fwrite(i, ary_file_fd);
		} else if (sa_get_ignored_chars() != NULL &&
			   strchr(sa_get_ignored_chars(), text[i]) != NULL) {
		    /* do nothing */
		} else if ((mki_mode & SA_ONLY_00_7F) &&
			   (unsigned char)text[i] >= 0x80) {
		    /* do nothing */
		} else if ((mki_mode & SA_ONLY_80_FF) &&
			   (unsigned char)text[i] < 0x80) {
		    /* do nothing */
		} else
		    sa_fwrite(i, ary_file_fd);		    

		pre_char_is_delimiter = 0;
	    }

	} else {
	    fprintf(stderr, "ERROR(sa_write_index()): specify sort mode.");
	    exit(2);
	}

    }
    
    (void)fclose(ary_file_fd);
    sa_close_mmap(mmr);

    return SUCCESS;
}



/*
 * Compare SA_INDEX-es.
 */
static int
longcomp(const SA_INDEX *i, const SA_INDEX *j)
{
    return *i - *j;
}


/*
 * Make a doc-id file.
 */
SA_STAT sa_make_did(const char *txt_fname, char *ary_fname, char* did_fname,
		    const char* tag1, const char* tag2, int mki_mode)
{
    SUFARY* ary;
    SUF_RESULT sr_tag1;
    SA_INDEX i;
    SA_INDEX num;
    SA_INDEX *rslt;
    FILE *of;

    char fname_tmp[SA_FILE_NAME_MAX];

    if (did_fname == NULL || did_fname[0] == '\0')
	did_fname = sa_add_suffix_to_file_name(fname_tmp, txt_fname, "did");

    if ((ary = sa_open(txt_fname, ary_fname)) == NULL) {
	sa_error_no = FILE_OPEN_ERROR;
	return FAIL;
    }

    sr_tag1 = sa_find(ary, 0, ary->arraysize - 1, tag1, strlen(tag1), 0);
    if (sr_tag1.stat == FAIL) {
	printf("pattern \"%s\" can not be found.\n", tag1);
	return FAIL;
    }

    num = sr_tag1.right - sr_tag1.left + 1;

    rslt = malloc(num * 2 * sizeof(SA_INDEX));
	
    if ((of = fopen(did_fname, "wb")) == NULL) {
	sa_error_no = FILE_OPEN_ERROR;
	return FAIL;
    }

    if (mki_mode & SA_VERBOSE)
	printf("Number of Documents = %ld\n", (long)num);

    for (i = 0; i < num; i++)
	rslt[i] = sa_aryidx2txtidx(ary, sr_tag1.left + i);

/* for (i = 0; i < num; i++) printf("p %ld\n", rslt[i]); */

    if (tag2 != NULL && tag2[0] != '\0') {
	/* ɥȽλꤵƤ */
	int len_tag2 = strlen(tag2);
	SUF_RESULT sr_tag2;

	sr_tag2 = sa_find(ary, 0, ary->arraysize - 1, tag2, strlen(tag2), 0);
	if (sr_tag2.stat == FAIL) { /* ɥȳϥ */
	    printf("pattern \"%s\" can not be found.\n", tag2);
	    return FAIL;
	}

	if(num != sr_tag2.right - sr_tag2.left + 1) {
	    printf("number of \"%s\" != number of \"%s\"\n", tag1, tag2);
	    return FAIL;
	}

	for (i = 0; i < num; i++)
	    rslt[num + i] = sa_aryidx2txtidx(ary, sr_tag2.left + i) + len_tag2;

	qsort(rslt, (size_t)(num * 2), sizeof(SA_INDEX),
	      (int(*)(const void*, const void*))longcomp);

    } else {
	/* ɥȽλ  ϥ */

	qsort(rslt, (size_t)num, sizeof(SA_INDEX),
	      (int(*)(const void*, const void*))longcomp);

	rslt[num * 2 - 1] = sa_get_text_size(ary);
	for (i = num * 2 - 2; i > 0; i -= 2)
	    rslt[i] = rslt[i - 1] = rslt[i / 2];

    }

    /* for (i = 0; i < num * 2; i++) printf("a %ld\n", rslt[i]); */

    for (i = 0; i < num * 2; i++)
	sa_fwrite(rslt[i], of);

    free(rslt);

    fclose(of);
    sa_close(ary);

    return SUCCESS;
}
