/*
 * $Id: sass.c,v 1.32 2000/02/06 11:09:20 tatuo-y Exp $
 *
 * sass --- Suffix Array Simple Search
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "sufary.h"
#include "getopt.h"

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#define TAG_LEN_MAX 1000
#define INPUT_STR_LEN_MAX 10000
#define FILE_NUM_MAX 256

int regex_mode = 0;
int stdin_mode = 0;
int ignore_case_mode = 0;
int before_context = 0;
int after_context = 0;
int keyword_emphasis_mode = 0;
int tag_mode = 0;
int uniq_mode = 0;
int sort_mode = 0;		/* sort by the position in the file */
char tags[2][TAG_LEN_MAX];
SA_INDEXES *rgnary;		/* for sort_mode */
SA_INDEX rgnary_idx = 0;
SA_INDEX size_of_rgnary;

static void search_all_files(char *fnames[], int num_of_files,
			     const char *key);
static void search_for_the_given_pattern(const SUFARY *ary, const char *key,
					 const char *fname);
static void show_usage(void);
static void show_mini_usage(void);
static int parse_options(int, char**);
static int ck_atoi(char const *, int *);

/*
 * Command line options.
 */
static const char *short_options = "0123456789hirsA:B:C::kvdt:Su";
static struct option long_options[] = {
    {"help",             no_argument,       NULL, 'h'},
    {"ignore-case",      no_argument,       NULL, 'i'},
    {"regex",            no_argument,       NULL, 'r'},
    {"stdin",            no_argument,       NULL, 's'},
    {"keyword-emphasis", no_argument,       NULL, 'k'},
    {"after-context",    required_argument, NULL, 'A'},
    {"before-context",   required_argument, NULL, 'B'},
    {"context",          optional_argument, NULL, 'C'},
    {"tags",             required_argument, NULL, 't'},
    {"debug",            no_argument,       NULL, 'd'},
    {"version",          no_argument,       NULL, 'v'},
    {"sort",             no_argument,       NULL, 'S'},
    {"uniq",             no_argument,       NULL, 'u'},
    {"0",                no_argument,       NULL, '0'},
    {"1",                no_argument,       NULL, '1'},
    {"2",                no_argument,       NULL, '2'},
    {"3",                no_argument,       NULL, '3'},
    {"4",                no_argument,       NULL, '4'},
    {"5",                no_argument,       NULL, '5'},
    {"6",                no_argument,       NULL, '6'},
    {"7",                no_argument,       NULL, '7'},
    {"8",                no_argument,       NULL, '8'},
    {"9",                no_argument,       NULL, '9'},
    {NULL, 0, NULL, 0}
};


int
main(int argc, char *argv[])
{
    int argc_offset;
    int num_of_files;
    char* key = NULL;
    argc_offset = parse_options(argc, argv);

    if (argc_offset == argc) {
	show_mini_usage();
	exit(1);
    }

    if (!stdin_mode) {
	key = argv[argc_offset];
	argc_offset++;
    }

    num_of_files = argc - argc_offset;
    if (num_of_files < 1) {
	show_mini_usage();
	exit(1);
    }

    if (stdin_mode) {
	char cmd[INPUT_STR_LEN_MAX];
	while (fgets(cmd, (int)sizeof(cmd), stdin)) {
	    cmd[strlen(cmd) - 1] = '\0'; /* delete '\n' */
	    search_all_files(argv + argc_offset, num_of_files, cmd);
	}
    } else {
	search_all_files(argv + argc_offset, num_of_files, key);
    }

    return 0;
}


static void
search_all_files(char *fnames[], int num_of_files, const char *key)
{
    SUFARY *ary[FILE_NUM_MAX];
    int i;
    for (i = 0; i < num_of_files; i++) {
	if ((ary[i] = sa_open(fnames[i], NULL)) == NULL) {
	    exit(1);
	}
	if (num_of_files == 1)
	    search_for_the_given_pattern(ary[i], key, NULL);
	else
	    search_for_the_given_pattern(ary[i], key, fnames[i]);
	sa_close(ary[i]);
    }
}


/*
 * Get region information and store tmporary array (for sorting).
 */
static SA_INDEXES
get_one_region(const SUFARY *ary, SA_INDEX pos)
{
    SA_INDEXES rgn;
    if (tag_mode) {
	rgn = sa_get_region_by_scan(ary, pos, tags[0], tags[1], 0, 0);
    } else {
	rgn = sa_get_region_by_scan(ary, pos, "\n", "\n",
				    before_context, after_context);
	if (*sa_txtidx2txtptr(ary, rgn.from) == '\n')
	    rgn.from++;
	if (*sa_txtidx2txtptr(ary, rgn.to - 1) == '\n')
	    rgn.to--;
    }

    if (sort_mode) {
	rgnary[rgnary_idx] = rgn;
	rgnary_idx++;
    }

    return rgn;
}


static void 
print_one_region(const SUFARY *ary, SA_INDEXES rgn, const char *fname,
		 int key_len)
{
    char* s;
    SA_INDEX rgn_len = rgn.to - rgn.from;
    s = sa_txtidx2txtptr(ary, rgn.from);

    if (fname != NULL)
	printf("%s:", fname);

    if (keyword_emphasis_mode) {
	int pos_in_rgn = rgn.here - rgn.from;
	printf("%.*s", pos_in_rgn, s);
	printf("\x1b[7m%.*s\x1b[0m", key_len, s + pos_in_rgn);
	printf("%.*s\n", (int)(rgn_len - (pos_in_rgn + key_len)),
	       s + pos_in_rgn + key_len);
    } else {
	printf("%.*s\n", (int)rgn_len, s);
    }

    if (after_context >= 1 || before_context >= 1 || tag_mode) {
	printf("--\n");
    }
}


static int
rgn_cmp(const void *data1, const void *data2)
{
    SA_INDEXES *a = (SA_INDEXES *)data1;
    SA_INDEXES *b = (SA_INDEXES *)data2;
    return a->from - b->from;
}


static void
print_sorted_result(const SUFARY *ary, const char *fname)
{
    int i;
    SA_INDEX pre = -1;
    qsort(rgnary, size_of_rgnary, sizeof(SA_INDEXES),
	  (int(*)(const void*,const void*))(rgn_cmp)); 
    for (i = 0; i < size_of_rgnary; i++) {
	if (uniq_mode == 0 || pre != rgnary[i].from)
	    print_one_region(ary, rgnary[i], fname, 2);
	pre = rgnary[i].from;
    }
}


/*
 * Change the key to a regex string: the -> [tT][hH][eE]
 */
static void
expand_ignore_case(char *new_key, const char *key)
{
    int i, j = 0;
    for (i = 0; i < strlen(key); i++) {
	if (isalpha(key[i])) {
	    new_key[j] = '[';
	    new_key[j + 1] = tolower(key[i]);
	    new_key[j + 2] = toupper(key[i]);
	    new_key[j + 3] = ']';
	    j += 4;
	} else if (key[i] == '[' || key[i] == ']' || key[i] == '\\' ||
		   key[i] == '(' || key[i] == ')' || key[i] == '.' ||
		   key[i] == '|' || key[i] == '*' || key[i] == '?') {
	    new_key[j] = '\\';
	    new_key[j + 1] = key[i];
	    j += 2;
	} else {
	    new_key[j] = key[i];
	    j++;
	}
    }
    new_key[j] = '\0';
}


/*
 * Find a keyword and print regions.
 */
static void
search_for_the_given_pattern(const SUFARY *ary, const char *key,
			     const char *fname)
{
    if (strlen(key) == 0)
	return;

    if (regex_mode || ignore_case_mode) {
	REGEX_RESULT_LIST *ll, *tmp;
	char *tmp_key = (char *)key;

	if (ignore_case_mode) {
	    tmp_key = malloc(strlen(key) * 4 + 1); /* estimation: strlen x 4 */
	    expand_ignore_case(tmp_key, key);
	}

        tmp = sa_regex(ary, 0, ary->arraysize - 1, tmp_key, strlen(tmp_key));

	if (sort_mode) {
	    size_of_rgnary = 0;
	    for (ll = tmp; ll != NULL; ll = ll->next)
		size_of_rgnary++;
	    rgnary = malloc(sizeof(SA_INDEXES) * size_of_rgnary);
	    assert(rgnary != NULL && "too many results!");
	}

        for (ll = tmp; ll != NULL; ll = ll->next) {
	    long pos = ll->value;
	    SA_INDEXES rgn;
	    rgn = get_one_region(ary, pos);
	    if (!sort_mode)
		print_one_region(ary, rgn, fname, ll->len);
        }

	if (ignore_case_mode) 
	    free(tmp_key); 
        free_regex_result_list(tmp);

    } else {
	SUF_RESULT sr;
	sr = sa_find(ary, 0, ary->arraysize - 1, (char *)key, strlen(key), 0);

	if (sr.stat == SUCCESS) {
	    SA_INDEX ai;
	    if (sort_mode) {
		size_of_rgnary = sr.right - sr.left + 1;
		rgnary = malloc(sizeof(SA_INDEXES) * size_of_rgnary);
		assert(rgnary != NULL && "too many results!");
	    }
	    for (ai = sr.left; ai <= sr.right; ai++) {
		SA_INDEX pos = sa_aryidx2txtidx(ary, ai);
		SA_INDEXES rgn = get_one_region(ary, pos);
		if (!sort_mode)
		    print_one_region(ary, rgn, fname, strlen(key));
	    }
	}
    }

    if (sort_mode) {
	print_sorted_result(ary, fname);
	rgnary_idx = 0;
	free(rgnary);
    }
}


/*
 * Parse tagstring: "<a>,</a>" --> "<a>" and "</a>"
 */
static void
parse_tags(const char *tgstr)
{
    int i, j;
    int tag_id = 0;
    for (i = 0, j = 0; i <= strlen(tgstr); i++, j++) {
	if (tgstr[i] == '\\' && i < strlen(tgstr) && tgstr[i + 1] == 'n') {
	    tags[tag_id][j] = '\n';
	    i++;
	} else if (tgstr[i] == ',') {
	    tags[tag_id][j] = '\0';
	    tag_id = 1;
	    j = -1;
	} else {
	    tags[tag_id][j] = tgstr[i];
	}
    }
/*     printf("[%s] -> [%s] [%s]\n", tgstr, tags[0], tags[1]); */
}


static void
show_usage(void)
{
    printf("\
sass %s, simple search program of SUFARY.\n\n\
Usage: sass [options] <pattern> <file>...\n\
  -r, --regex               use regex for the pattern.\n\
  -i, --ignore-case         ignore case.\n\
  -s, --stdin               read the pattern from stdin.\n\
  -k, --keyword-emphasis    emphasize keyword in results\n\
  -A, --after-context=NUM   print NUM lines of trailing context\n\
  -B, --before-context=NUM  print NUM lines of leading context\n\
  -C, --context[=NUM]       print NUM (default 2) lines of output context\n\
                            unless overriden by -A or -B\n\
  -NUM                      same as --context=NUM\n\
  -t, --tags=TAG1,TAG2      print text regions (from TAG1 to TAG2).\n\
                            example: -t '<doc>,</doc>', -t '\\n\\n,\\n\\n'\n\
  -S, --sort                sort results by the position in the file.\n\
  -u, --uniq                similar to --sort, but\n\
                            only output the first of an equal sequence.\n\
  -v, --version             show the version of SUFARY and exit.\n\
  -d, --debug               show the debug information.\n\n\
Notes:\n\
  You should prepare an array file for each target file.\n\
  (these file names must be its original file name followed by \".ary\")\n",
	   VERSION);
}


static void
show_mini_usage(void)
{
    printf("\
Usage: sass [options] <pattern> <file ...>\n\
Try `sass --help' for more information.\n");
}


static int
parse_options(int argc, char **argv)
{
    int digit_args_val = 0;

    while (1) {
        int ch = getopt_long(argc, argv, short_options, long_options, NULL);
        if (ch == EOF) {
            break;
	}
	switch (ch) {
	case 'h':
	    show_usage();
	    exit(1);
	    break;
	case 'i':
	    ignore_case_mode = 1;
	    break;
	case 'r':
	    regex_mode = 1;
	    break;
	case 's':
	    stdin_mode = 1;
	    break;
	case 'A':
	    if (ck_atoi(optarg, &after_context)) {
		fprintf(stderr, "invalid context length argument\n");
		exit(1);
	    }
	    break;
	case 'B':
	    if (ck_atoi(optarg, &before_context)) {
		fprintf(stderr, "invalid context length argument\n");
		exit(1);
	    }
	    break;
	case 'C':
	    if (optarg) {
		if (ck_atoi(optarg, &before_context)) {
		    fprintf(stderr, "invalid context length argument\n");
		    exit(1);
		}
		after_context  = before_context;
	    } else {
		before_context = 2;
		after_context  = 2;
	    }

	    break;
	case 'd':
	    sa_set_debug_mode(1);
	    break;
	case 't':
	    tag_mode = 1;
	    parse_tags(optarg);
	    break;
	case 'k':
	    keyword_emphasis_mode = 1;
	    break;
	case 'u':
	    uniq_mode = sort_mode = 1;
	    break;
	case 'S':
	    sort_mode = 1;
	    break;
	case 'v':
            printf("sass of SUFARY %s\n", VERSION);
	    printf("Copyright (C) 1997-1999 YAMASHITA Tatuo All rights reserved.\n");
            exit(1);
	    break;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    digit_args_val = 10 * digit_args_val + ch - '0';
	    before_context = digit_args_val;
	    after_context  = digit_args_val;
	    break;
	default:
	    show_mini_usage();
	    exit(1);
	}
    }
    return optind;
}

/* Imported from GNU grep-2.3 [1999-11-08] by satoru-t */

/* Convert STR to a positive integer, storing the result in *OUT.
   If STR is not a valid integer, return -1 (otherwise 0). */
static int
ck_atoi (str, out)
     char const *str;
     int *out;
{
  char const *p;
  for (p = str; *p; p++)
    if (*p < '0' || *p > '9')
      return -1;

  *out = atoi (optarg);
  return 0;
}
